/* This file is part of the model extraction module to FaCT++
Copyright (C) 2003-2008 by Johannes Bauer

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#include "ModelBuilding.h"
#include <unistd.h>
#include <iostream>

ModelBuilderCPP :: ModelBuilderCPP(JNIEnv * env, jobject builder, jobject instance) {
	indivCount = 0;
	this->env = env;
	this->builder = builder;
	this->instance = instance;
	exc = NULL;
	modelBuilderJ = env->GetObjectClass(builder);
	jclass exceptionclass = env->FindClass("java/lang/RuntimeException");
	pushObjectPropertyID = env->GetMethodID(modelBuilderJ, "pushObjectProperty", "(Ljava/lang/String;Ljava/lang/String;)V");
	if(pushObjectPropertyID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method pushObjectProperty."); 
		return;
	}
	popObjectPropertyID = 	env->GetMethodID(modelBuilderJ, "popObjectProperty", "()V");
	if(popObjectPropertyID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method popObjectProperty."); 
		return;
	}
	addConceptID = 	env->GetMethodID(modelBuilderJ, "addConcept", "(Ljava/lang/String;)V");
	if(addConceptID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method addConcept."); 
		return;
	}
	addNegatedConceptID = 	env->GetMethodID(modelBuilderJ, "addNegatedConcept", "(Ljava/lang/String;)V");
	if(addConceptID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method addNegatedConcept."); 
		return;
	}
	pruneID = 	env->GetMethodID(modelBuilderJ, "prune", "()V");
	if(pruneID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method prune."); 
		return;
	}
	initModelID = 	env->GetMethodID(modelBuilderJ, "initModel", "(Ljava/lang/String;)V");
	if(initModelID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method initModel."); 
		return;
	}
	aboxInconsistentID = 	env->GetMethodID(modelBuilderJ, "aboxInconsistent", "()V");
	if(aboxInconsistentID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method aboxInconsistent."); 
		return;
	}
	conceptMandatoryID = 	env->GetMethodID(modelBuilderJ, "conceptMandatory", "()V");
	if(conceptMandatoryID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method conceptMandatory."); 
		return;
	}
	roleMandatoryID = 	env->GetMethodID(modelBuilderJ, "roleMandatory", "()V");
	if(roleMandatoryID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method roleMandatory."); 
		return;
	}
	setBlockerID = 	env->GetMethodID(modelBuilderJ, "setBlocker", "(Ljava/lang/String;)V");
	if(setBlockerID == 0) { 
		env->ThrowNew(exceptionclass, "Could not find method setBlocker."); 
		return;
	}
}

ModelBuilderCPP :: ~ModelBuilderCPP() {
	for(map<const TConcept*, set<const TConcept*>*>::iterator alpha = synMap.begin(),	omega = synMap.end();
			alpha != omega;
			++alpha) {
			set<const TConcept*>* delSet = alpha->second;
			synMap[alpha->first] = NULL;
			delete(delSet);
	}
}

void ModelBuilderCPP :: buildModel(const DlCompletionTree* tree, int depth) {
	if(getException() != NULL)
		return;
	
	const DlCompletionTree* blocker = tree->getDBlocker();
	if(blocker)
		setBlocker(blocker);

	if(getStatus(tree) == UNSEEN) {
		setStatus(tree, EXPLORING);

			/// Collect Types
		map<set<const TConcept*>*, short> label;
		for(DlCompletionTree::const_label_iterator
					 alpha = tree->beginl_sc(),
					 omega = tree->endl_sc();
				alpha != omega;
				++alpha)
			addConcept(alpha->bp(), alpha->getDep().empty(), label);

			/// Add Types
		for(map<set<const TConcept*>*, short>::const_iterator alpha = label.begin(), omega = label.end();
			alpha != omega;
			++alpha) {
			assert(alpha->first != NULL);
			for(set<const TConcept*>::const_iterator beta = alpha->first->begin(), gamma = alpha->first->end();
					beta != gamma;
					++beta) {
				if(alpha->second > 1)
					addConcept((*beta)->getName());
				else
					addNegatedConcept((*beta)->getName());
				if(alpha->second % 2)
					conceptMandatory();
			}
		}


	
		DlCompletionTree::const_edge_iterator p, p_end;
		const TRole* role;
		bool rMand;
	  DlCompletionTree* object;
		depth--;
		for ( p = tree->begins(), p_end = tree->ends(); p != p_end && getException() == NULL ; ++p ) {
			role = (*p)->getRole();
			if(role == NULL)	// object was P-Blocked, purged.
				continue;
			rMand = (*p)->getDep().empty();
			object = (*p)->getArcEnd();
	    assert(role != NULL);
	    assert(object != NULL);
			jstring objectName = getNameFor(object);
			jstring subjectName = getNameFor(tree);

			if(role->getId() > 0) {
				pushObjectProperty(role->getName(), rMand, objectName);
				// add inverse role
				const TRole* inv = role->inverse();
				if(inv->getId() > 0) {
					pushObjectProperty(inv->getName(), rMand, subjectName);
					popObjectProperty();
				}
					// add inverse super roles
				for(TRole::roleSet::const_iterator beta = inv->begin_anc(), psi = inv->end_anc();
					beta != psi;
					++beta) {
					if((*beta)->getId() > 0) {
						pushObjectProperty((*beta)->getName(), rMand, subjectName);
						popObjectProperty();
					}
				}
				if(depth < 1 && role->getId() > 0)
					prune(object);
				else {
					short objStat = getStatus(object);	
					if(objStat != EXPLORED && objStat != EXPLORING)
						buildModel(object, depth);
					popObjectProperty();
				}
					// add super roles
				for(TRole::roleSet::const_iterator beta = role->begin_anc(), psi = role->end_anc();
					beta != psi;
					++beta) {
					if((*beta)->getId() > 0) {
						pushObjectProperty((*beta)->getName(), rMand, objectName);
						popObjectProperty();
					}
				}
			}
		}
		setStatus(tree, EXPLORED);
	}
}

void ModelBuilderCPP :: addConcept(BipolarPointer bp, bool mandatory, map<set<const TConcept*>*, short> &label) {
	DLVertex* v = &(*DLHeap)[bp];
	switch (v->Type()) {
	case dtPConcept:
	case dtNConcept: {
		short mand = 0;
		if(mandatory)
			mand = 1;
	
		short pos = 0;
		if(isPositive(bp))
			pos = 2;
			
		const TConcept* theConcept = (const TConcept*) v->getConcept();
		set<const TConcept*> *synonyms = synMap[theConcept];
		if(synonyms == NULL) { // can't happen, can it?
			synonyms = new set<const TConcept*>();
			synonyms->insert(theConcept);
			synMap[theConcept] = synonyms;
		}
		map<set<const TConcept*>*,short>::iterator s = label.find(synonyms);
		if(s == label.end() || (mand && !(s->second & 1)))
			label[synonyms] = mand + pos;
		break;
	}
	case dtAnd: {
		if(isPositive(bp)) { // Otherwise we could start a disjunction.
			DLVertex::const_iterator alpha, omega;
			for(alpha = v->begin(), omega = v->end(); alpha != omega; ++alpha)
				addConcept(*alpha, mandatory, label);
		}
	}
	default:
		break;
	}
}

short ModelBuilderCPP :: getStatus(const DlCompletionTree* tree) {
	pair<jstring,short> *p = seenNodes[tree];
	if(p == NULL)
		return UNSEEN;

	return p->second;
}

void ModelBuilderCPP :: setStatus(const DlCompletionTree* tree, short status) {
		// make sure the node is known.
	getNameFor(tree);

	pair<jstring,short> *p = seenNodes[tree];
	p->second = status;
}

jstring ModelBuilderCPP :: getNameFor(const DlCompletionTree* tree) {
  assert(tree != NULL);
	pair<jstring,short>* p = seenNodes[tree];
	if(p == NULL) {
		jstring jname = NULL;
		for(DlCompletionTree::const_label_iterator
					 alpha = tree->beginl_sc(),
					 omega = tree->endl_sc();
				alpha != omega && jname == NULL;
				++alpha)
		{
			if ( isPositive(alpha->bp()) )
			{
					DLVertex* v = &(*DLHeap)[alpha->bp()];
	
					if ( v->Type() == dtPSingleton ||  v->Type() == dtNSingleton) {
						jname = env->NewStringUTF(v->getConcept()->getName());
						break;
					}
			}
		}
		if(jname == NULL) {
			std::stringstream strstr;
			do {
				indivCount++;
				strstr.str("");
				strstr << std::hex << indivCount;
			} while (getK(env, instance)->isSingletonName(strstr.str()));
			jname = env->NewStringUTF(strstr.str().c_str());
		}
		p = new pair<jstring,short>;
		p->first = jname,
		p->second = UNSEEN;
		seenNodes[tree] = p;
	}
	return p->first;
}

void ModelBuilderCPP :: setupSynonyms(const ReasoningKernel *kernel) {
	for(TBox::c_const_iterator alpha = kernel->c_begin(), omega = kernel->c_end() ;
			alpha != omega;
			++alpha) {

		const TConcept *syn;
		if((*alpha)->isSynonym())
			syn = (TConcept*) (*alpha)->getSynonym();
		else
			syn = *alpha;

		set<const TConcept*> *syns = synMap[syn];
		if(syns == NULL) {
			syns = new set<const TConcept*>();
			synMap[syn] = syns;
		}
		syns->insert(*alpha);
	}
}

void ModelBuilderCPP :: noteException(const char* exc) {
	this->exc = exc;
	env->ExceptionDescribe();
	env->ExceptionClear();
}

const char* ModelBuilderCPP :: getException() {
	return exc;
}

///////////////////
// java method wrappers
///////////////////

void ModelBuilderCPP :: initModel(const DlCompletionTree* tree, const ReasoningKernel* kernel, DLDag* DLHeap, int depth) {
	assert(env != NULL);
	this->DLHeap = DLHeap;
	jthrowable exc;
	jstring name = getNameFor(tree);
	env->CallVoidMethod(builder, initModelID, name);
	exc = env->ExceptionOccurred();
	if(exc) {
		env->ExceptionDescribe();
		env->ExceptionClear();
		jclass exClass = env->FindClass("uk/ac/manchester/cs/owl/supermodel/SuperModelReasonerException");
		env->ThrowNew(exClass, "Thrown in builder.initModel.");
	}
	setupSynonyms(kernel);
	buildModel(tree, depth);
}

void ModelBuilderCPP :: pushObjectProperty(const char* name, bool mand, jstring object) {
	jstring nameStr = env->NewStringUTF(name);
	env->CallVoidMethod(builder, pushObjectPropertyID, nameStr, object);
	if(mand)
		roleMandatory();
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during pushObjectProperty.");
}

void ModelBuilderCPP :: popObjectProperty() {
	env->CallVoidMethod(builder, popObjectPropertyID);
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during popObjectProperty.");
}

void ModelBuilderCPP :: addConcept(const char* name) {
	jstring nameStr = env->NewStringUTF(name);
	env->CallVoidMethod(builder, addConceptID, nameStr);
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during addConcept.");
}

void ModelBuilderCPP :: addNegatedConcept(const char* name) {
	jstring nameStr = env->NewStringUTF(name);
	env->CallVoidMethod(builder, addNegatedConceptID, nameStr);
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during addNegatedConcept.");
}

void ModelBuilderCPP :: prune(const DlCompletionTree* tree) {
	env->CallVoidMethod(builder, pruneID);
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during prune.");
	setStatus(tree, PRUNED);
}

void ModelBuilderCPP :: aboxInconsistent() {
	env->CallVoidMethod(builder, aboxInconsistentID);
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during aboxInconsistent.");
}

void ModelBuilderCPP :: conceptMandatory() {
	env->CallVoidMethod(builder, conceptMandatoryID);
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during conceptMandatory.");
}

void ModelBuilderCPP :: roleMandatory() {
	env->CallVoidMethod(builder, roleMandatoryID);
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during roleMandatory.");
}
void ModelBuilderCPP :: setBlocker(const DlCompletionTree* blocker) {
	env->CallVoidMethod(builder, setBlockerID, getNameFor(blocker));
	if(env->ExceptionOccurred())
		noteException("Exception occurred on the java side during setBlocker.");
}
